#include "grammar.h"
#include "tools.h"
#include <string.h>     /* memcpy(), strcmp() */
#include <assert.h>
#include <stdlib.h>     /* free() */



/******************************
 *                            *
 *                            *
 *    Methods for Alphabet    *
 *                            *
 *                            *
 ******************************/

Alphabet *alphabet_new(void)
{
    Alphabet *alphabet = (Alphabet *) xmalloc(sizeof(Alphabet));
    alphabet->head = alphabet->tail = NULL;
    return alphabet;
}

/*
 * Adds new symbol to the combined terminal/nonterminal alphabet.  If a symbol
 * with the given name already exists, checks that it's type (terminal /
 * nonterminal) is the same.  The argument `symb_ret', if not NULL, is used to
 * pass back the pointer to Symb struct.
 *
 * Return values:
 *
 *     0   type mismatch error. Name was not added because the name is already
 *         in the alphabet and has type (terminal / nonterminal) not matching
 *         requested type (term).  Pointer to the matching symbol is stored in
 *         symb_ret;
 *
 *     1   success.  Name was not added because the name is already in the
 *         alphabet, and has the same type.  Pointer to the matching symbol is
 *         stored in symb_ret;
 *
 *     2   success.  Name was added to the alphabet.  Pointer to the created
 *         symbol is stored in symb_ret.
 */
int alphabet_add_name(
    Alphabet *alphabet, int terminal, char *name, int len, Symb **symb_ret)
{
    Symb *symb;

    assert(alphabet);
    assert(name);
    assert(len>=0);
    terminal = terminal ? 1 : 0;    /* restrict to 0 and 1 */

    if ( alphabet->head ) {         /* alphabet not empty */
        /* Try and lookup the name in the alphabet */
        symb = alphabet_find_name(alphabet, name);
        if ( symb ) {               /* name already in the alphabet */
            if ( symb_ret ) *symb_ret = symb;
            if ( symb->terminal != terminal ) /* type mismatch */
                return 0;
            return 1;   /* type matched up, just pass back the pointer */
        } else {        /* name is not in the alphabet, create new symbol */
            alphabet->tail->next = (Symb *) xmalloc(sizeof(Symb));
            symb = alphabet->tail = alphabet->tail->next;
        }
    } else              /* alphabet empty, initialize it */
        symb = alphabet->head = alphabet->tail = (Symb *) xmalloc(sizeof(Symb));

    /* Initialize the new symbol */
    symb->name = (char *) xmalloc(len+1);   /* include trailing '\0' */
    memcpy( symb->name, name, len+1 );      /* include trailing '\0' */
    symb->terminal = terminal;
    symb->next = NULL;                      /* adding to the end of the list */

    if ( symb_ret ) *symb_ret = symb;       /* pass back the created struct */

    return 2;
}

/* Searches the alphabet for a symbol with the given name.  Returns pointer to
 * it, or NULL if such symbol not found */
Symb *alphabet_find_name(Alphabet *alphabet, char *name)
{
    Symb *symb = alphabet->head;

    assert(alphabet);
    assert(name);

    while ( symb ) {
        if ( !strcmp(symb->name, name) )
            break;
        symb = symb->next;
    }

    return symb;
}

/* Searches the alphabet for a symbol.  Returns pointer to PREVIOUS symbol, or
 * NULL if such symbol not found */
Symb *alphabet_find_symb(Alphabet *alphabet, char *name)
{
    Symb *symb = alphabet->head;

    assert(alphabet);
    assert(name);

    while ( symb ) {
        if ( !strcmp(symb->name, name) )
            break;
        symb = symb->next;
    }

    return symb;
}

/* Removes from the alphabet all symbols which are in the given string (set) */
void alphabet_remove_set(Alphabet *alphabet, SItem *set)
{
    Symb *symb, *prev;

    assert(alphabet);

    if ( !set ) /* nothing to remove */
        return;

    symb = alphabet->head;
    prev = NULL;
    while ( symb ) {
        if ( string_find_sitem(set, symb) ) {
            /* The symbol from the alphabet is in the set, remove it */
            Symb *temp = symb->next; /* save for later reference */
            if ( !prev )    /* this is head symbol of the alphabet */
                alphabet->head = temp;
            else            /* this is not head symbol of the alphabet */
                prev->next = temp;
            free(symb->name);
            free(symb);
            symb = temp;
        } else {
            /* The symbol from the alphabet is not in the set, skip it */
            prev = symb;
            symb = symb->next;
        }
    }

    return;
}



/******************************
 *                            *
 *                            *
 *     Methods for Grammar    *
 *                            *
 *                            *
 ******************************/

Grammar *grammar_new(void)
{
    Grammar *grammar  = (Grammar *) xmalloc(sizeof(Grammar));
    grammar->alphabet = NULL;
    grammar->start    = NULL;
    grammar->head     = NULL;
    return grammar;
}

/* Makes a copy of a grammar.  Note that the copy shares the alphabet with the
 * original grammar (alphabet is not copied) */
Grammar *grammar_copy(Grammar *grammar)
{
    Grammar *grammar_out;
    Prod    *curr, *prev, *copy;

    assert(grammar);

    grammar_out = (Grammar *) xmalloc(sizeof(Grammar));
    grammar_out->alphabet = grammar->alphabet;
    grammar_out->start    = grammar->start;

    if ( !grammar->head ) {    /* empty productions list, all done */
        grammar_out->head = NULL;
        return grammar_out;
    }

    /* Make copy of the first production */
    grammar_out->head = production_copy(grammar->head);
    prev = grammar_out->head;
    copy = grammar->head->next;

    while ( copy ) {
        curr = production_copy(copy);
        prev->next = curr;
        prev = curr;
        copy = copy->next;
    }

    prev->next = NULL; /* close the list */

    return grammar_out;
}

/* Creates a new production with the given head symbol (at the end of
 * productions list).  Before doing so, checks that such production is not
 * already in the grammar; if it is, just returns pointer to the existing
 * production */
Prod *grammar_add_empty_production(Grammar *grammar, Symb *head)
{
    Prod *prod;

    assert(grammar);
    assert(head);
    assert(!head->terminal);

    if ( grammar->head ) { /* grammar not empty */
        prod = grammar->head;
        /* See if a production with this head already exists */
        while ( prod->next && prod->hsymb != head )
            prod = prod->next;
        if ( prod->hsymb == head )   /* found it */
            return prod;
        /* No production with this head, add new one at the end of the list  */
        assert(!prod->next);        /* should be tail production */
        prod->next = (Prod *) xmalloc(sizeof(Prod));
        prod = prod->next;
    } else  /* grammar empty, initialize it */
        prod = grammar->head = (Prod *) xmalloc(sizeof(Prod));

    /* Fill in the new production */
    prod->next  = NULL;  /* we add to the end of list */
    prod->hsymb = head;
    prod->head  = NULL;

    return prod;
}

/* Searches the grammar for a production whose head is a symbol with the given
 * name.  Returns pointer to it, or NULL if such production not found */
Prod *grammar_find_production_by_name(Grammar *grammar, char *name)
{
    Prod *prod;

    assert(grammar);

    prod = grammar->head;
    while ( prod ) {
        if ( !strcmp(prod->hsymb->name, name) )
            break;
        prod = prod->next;
    }

    return prod;
}

/* Searches the grammar for a production whose head is the given symbol.
 * Returns pointer to it, or NULL if such production not found */
Prod *grammar_find_production_by_symb(Grammar *grammar, Symb *head)
{
    Prod *prod;

    assert(grammar);

    prod = grammar->head;
    while ( prod ) {
        if ( prod->hsymb == head )
            break;
        prod = prod->next;
    }

    return prod;
}

void grammar_print(Grammar *grammar, FILE *stream)
{
    Prod  *prod;
    Symb  *symb;
    int print_token_def = 1;

    assert(grammar);
    assert(grammar->alphabet);
    if ( !stream )
        stream = stdout;

    /* Print out `%start' line */
    assert(grammar->start);
    assert(!grammar->start->terminal);
    fprintf(stream, "%%start %s\n", grammar->start->name);

    /* Print out `%token' line, if needed.  We assume that none of named token
     * types starts with `'', and that anything that starts with it is a
     * character token type */
    symb = grammar->alphabet->head;
    while ( symb ) {
        if ( *symb->name != '\'' && symb->terminal ) {
            if ( print_token_def ) {
                fputs("%token", stream);
                print_token_def = 0;
            }
            fprintf(stream, " %s", symb->name);
        }
        symb = symb->next;
    }
    if ( !print_token_def ) /* we have printed `%token ...' */
        fputc('\n', stream);/* ...so print a new line */
    fputc('\n', stream);    /* separate productions from the header */

    /* Print production blocks */
    prod = grammar->head;
    while ( prod ) {
        production_print(prod, stream);
        prod = prod->next;
    }

    return;
}



/******************************
 *                            *
 *                            *
 *      Methods for Prod      *
 *                            *
 *                            *
 ******************************/

/* Note that in the copy, next field is set to NULL */
Prod *production_copy(Prod *prod)
{
    Prod *prod_out;
    Body *curr, *prev, *copy;

    assert(prod);

    prod_out = (Prod *) xmalloc(sizeof(Prod));
    prod_out->hsymb = prod->hsymb;
    prod_out->next = NULL;

    if ( !prod->head ) {    /* empty bodies list, all done */
        prod_out->head = NULL;
        return prod_out;
    }

    /* Make copy of the first body */
    prod_out->head = body_copy(prod->head);
    prev = prod_out->head;
    copy = prod->head->next;

    /* Make copies of bodies that follow */
    while ( copy ) {
        curr = body_copy(copy);
        prev->next = curr;
        prev = curr;
        copy = copy->next;
    }

    prev->next = NULL;  /* close the list */

    return prod_out;
}

/* Adds an empty body to the given production (to the head of bodies list) */
Body *production_add_empty_body(Prod *prod)
{
    Body *body;

    assert(prod);

    body = (Body *) xmalloc(sizeof(Body));
    body->head = body->tail = NULL;
    body->next = prod->head;    /* add to the head of linked list */
    prod->head = body;

    return body;
}

Body *production_find_body(Prod *haystack, Body *needle)
{
    Body *body;

    assert(haystack);
/*     assert(haystack->head); */
    assert(needle);

    body = haystack->head;
    while ( body ) {
        if ( !string_cmp(body->head, needle->head) )
            break;
        body = body->next;
    }

    return body;
}

void production_print(Prod *prod, FILE *stream)
{
    Body *body;

    assert(prod);
    assert(prod->head);

    if ( !stream )
        stream = stdout;

    fprintf(stream, "%s:\t ", prod->hsymb->name);
    body = prod->head;
    /* Print bodies inside production block */
    while ( body ) {
        string_print(body->head, stream);
        body = body->next;
        if ( body ) /* not the end yet, need pipe */
            fputs("\n\t|", stream);
    }
    fputs("\n;\n\n", stream); /* finish off the production */

    return;
}

/*
 * This unlinks a production from grammar's chain list of productions and frees
 * the production and its bodies.  grammar must point to grammar this production
 * is from; prev_prod must be NULL for a first production in the chain list, or
 * point to previous production in that list otherwise.
 *
 * Return value is next production in the list, or NULL if curr_prod was the
 * tail
 */
Prod *production_unlink(Grammar *grammar, Prod *curr_prod, Prod *prev_prod)
{
    Prod *prod;

    assert(curr_prod);
    prod = curr_prod->next; /* save for later reference */

    production_free(curr_prod);
    if ( !prev_prod ) /* we are removing first production in the list */
        grammar->head = prod;
    else              /* this is not first production in the list */
        prev_prod->next = prod;

    return prod;
}

/* Note that this method does not update linked list this production is in, it
 * just frees the associated bodies and then frees itself */
void production_free(Prod *prod)
{
    Body *curr, *next;

    if ( !prod )
        return;

    curr = prod->head;
    while (curr) {
        next = curr->next;
        body_free(curr);
        curr = next;
    }

    free(prod);

    return;
}


/******************************
 *                            *
 *                            *
 *      Methods for Body      *
 *                            *
 *                            *
 ******************************/

/* Note that in the copy, next field is set to NULL */
Body *body_copy(Body *body)
{
    Body *body_out;

    if ( !body ) /* empty body, all done */
        return NULL;

    body_out = (Body *) xmalloc(sizeof(Body));
    body_out->next = NULL;
    body_out->head = string_copy(body->head, &body_out->tail);

    return body_out;
}

/* Adds an SItem for the given symbol to the end of body's string */
SItem *body_add_symb(Body *body, Symb *symb)
{
    SItem *sitem;

    assert(body);
    assert(symb);

    sitem = (SItem *) xmalloc(sizeof(SItem));
    sitem->symb = symb;
    sitem->next = NULL;     /* adding to the end of the list */
    if ( body->head ) {     /* string not empty */
        assert(body->tail);
        body->tail->next = sitem;
    }
    else {                  /* string empty, initialize it */
        assert(!body->tail);
        body->head = sitem;
    }

    body->tail = sitem;

    return sitem;
}

/*
 * This unlinks a body from production's chain list of bodies and frees the body
 * and its string.  prod must point to production this body is from; prev_body
 * must be NULL for a first body in the chain list, or point to previous body in
 * that list otherwise.
 *
 * Return value is next body in the list, or NULL if curr_body was the tail
 */
Body *body_unlink(Prod *production, Body *curr_body, Body *prev_body)
{
    Body *body;

    assert(curr_body);
    body = curr_body->next; /* save for later reference */

    body_free(curr_body);
    if ( !prev_body ) /* we are removing first body in the list */
        production->head = body;
    else              /* this is not first body in the list */
        prev_body->next = body;

    return body;
}

/* Note that this method does not update linked list this body is in, it just
 * frees the associated string and then frees itself */
void body_free(Body *body)
{
    if ( !body )
        return;
    string_free(body->head);
    free(body);
    return;
}



/******************************
 *                            *
 *                            *
 *    Methods for strings     *
 *                            *
 *                            *
 ******************************/

/* If tail_ret is not NULL, stores there pointer to new string's tail item */
SItem *string_copy(SItem *string, SItem **tail_ret)
{
    SItem *string_out = NULL, *curr, *prev;

    if ( !string ) { /* empty string, all done */
        if ( tail_ret ) *tail_ret = NULL;
        return NULL;
    }

    /* Make copy of the first item */
    prev = string_out = (SItem *) xmalloc(sizeof(SItem));
    prev->symb = string->symb;
    string     = string->next;

    /* Make copies of items that follow */
    while ( string ) {
        curr       = (SItem *) xmalloc(sizeof(SItem));
        prev->next = curr;
        curr->symb = string->symb;
        prev       = curr;
        string     = string->next;
    }

    prev->next = NULL; /* close the list */

    if ( tail_ret ) *tail_ret = prev; /* pass back the tail */

    return string_out;
}

/* Searches the given string (`haystack') for an item with the given symbol
 * (`needle').  Returns pointer to it, or NULL if no such item found */
SItem *string_find_sitem(SItem *haystack, Symb *needle)
{
/*     assert(haystack); */
    if ( !haystack )
        return NULL;

    while ( haystack ) {
        if ( haystack->symb == needle )
            break;
        haystack = haystack->next;
    }

    return haystack;
}

/* Searches the given string (`haystack') for an item with any symbol from the
 * given string (`needles').  Returns pointer to it, or NULL if no such item
 * found */
SItem *string_find_sitem_from_set(SItem *haystack, SItem *needles)
{
    SItem *item;

    assert(haystack);

    /* For each `needle', see if it is in the `haystack' */
    while ( needles ) {
        if ( (item = string_find_sitem(haystack, needles->symb)) )
            return item;
        needles = needles->next;
    }

    return NULL;
}

/* Returns 0 if two strings are equivalent, 1 if not */
int string_cmp(SItem *str1, SItem *str2)
{
    while ( str1 && str2 ) {
        if ( str1->symb != str2->symb )
            break;

        str1 = str1->next;
        str2 = str2->next;
    }

    if ( !str1 && !str2 )
        return 0;

    return 1;
}

/* Prints string */
void string_print(SItem *string, FILE *stream)
{
    if ( !stream )
        stream = stdout;

    while ( string ) {
        fprintf(stream, " %s", string->symb->name);
        string = string->next;
    }

    return;
}

void string_free(SItem *string)
{
    SItem *next;

    while ( string ) {
        next = string->next;
        free(string);
        string = next;
    }

    return;
}



/*
 * Local variables:
 * compile-command: "make -k && ./unreachable < grammar.3"
 * End:
 */
