#ifndef __GRAMMAR_H__
#define __GRAMMAR_H__


#include <stdio.h>  /* FILE */


/*
 * Alphabet is a singly linked list of symbols.  Each symbol can be either
 * terminal or nonterminal, so this alphabet is actually a union of terminal and
 * nonterminal alphabets
 */

/* These will form a singly linked list of symbols */
typedef struct _Symb {
    int   terminal;     /* 0 - nonterminal, 1 - terminal */
    char *name;         /* name of the symbol */
    struct _Symb *next; /* pointer to next symbol in the list */
} Symb;

/* This is alphabet struct proper */
typedef struct _Alphabet {
    Symb *head;
    Symb *tail;         /* for performance reasons when adding symbols */
} Alphabet;



/*
 * Grammar is a singly linked list of productions.  Each production actually
 * stands for a block of productions with common head symbol.  Therefore each
 * production contains not one body, but instead a singly linked list of bodies.
 * Each body contains pointer to head of a string (singly linked list of string
 * items).  Each item in that list contains pointer to the original symbol in
 * alphabet (see Symb struct above):
 *
 *        -- B - I - I - ... - I
 *       |
 *    -- P - B - I - I - ... - I
 *   |   |  ...
 *   |    -- B - I - I - ... - I
 *   |
 *   |
 *   |    -- B - I - I - ... - I
 *   |   |
 *   G - P - B - I - I - ... - I
 *   |   |  ...
 *   |    -- B - I - I - ... - I
 *   |
 *   |  ...
 *   |
 *   |
 *   |    -- B - I - I - ... - I
 *   |   |
 *    -- P - B - I - I - ... - I
 *       |  ...
 *        -- B - I - I - ... - I
 */

/* This is an item of a string, which refers to a symbol (terminal or
 * nonterminal).  String items are arranged in a singly linked list to form the
 * string */
typedef struct _SItem {
    Symb          *symb;/* (non)terminal symbol this item refers to */
    struct _SItem *next;/* pointer to next item in the string */
} SItem;

/* This is a body inside a block of productions.  Bodies are arranged in a
 * singly linked list representing all productions with common head symbol */
typedef struct _Body {
    SItem        *head; /* pointer to first symbol of the body */
    SItem        *tail; /* pointer to last symbol of the body */
    struct _Body *next; /* pointer to next body in the list */
} Body;

/* This is a block of all productions with common head symbol.  These blocks are
 * arranged in a singly linked list to form grammar */
typedef struct _Prod {
    Symb         *hsymb;/* production's head symbol (a nonterminal symbol) */
    Body         *head; /* singly linked list of production bodies */
    struct _Prod *next; /* pointer to next block of productions */
} Prod;

/* This is grammar struct proper */
typedef struct _Grammar {
    Alphabet *alphabet;
    Symb *start;        /* starting symbol from the alphabet */
    Prod *head;         /* head of the list of grammar's productions */
} Grammar;



/* Alphabet methods */
extern Alphabet *alphabet_new(void);
extern int       alphabet_add_name(
    Alphabet *alphabet, int terminal, char *name, int len, Symb **symb_ret);
extern Symb     *alphabet_find_name(Alphabet *alphabet, char *name);
extern void      alphabet_remove_set(Alphabet *alphabet, SItem *set);
/* Grammar methods */
extern Grammar  *grammar_new(void);
extern Grammar  *grammar_copy(Grammar *grammar);
extern Prod     *grammar_add_empty_production(Grammar *grammar, Symb *head);
extern Prod     *grammar_find_production_by_name(Grammar *grammar, char *name);
extern Prod     *grammar_find_production_by_symb(Grammar *grammar, Symb *head);
extern void      grammar_print(Grammar *grammar, FILE *stream);
/* Production methods */
extern Prod     *production_copy(Prod *prod);
extern Body     *production_add_empty_body(Prod *prod);
extern Body     *production_find_body(Prod *haystack, Body *needle);
extern void      production_print(Prod *prod, FILE *stream);
extern Prod     *production_unlink(Grammar *grammar, Prod *curr_prod, Prod *prev_prod);
extern void      production_free(Prod *prod);
/* Production body methods */
extern Body     *body_copy(Body *body);
extern SItem    *body_add_symb(Body *body, Symb *symb);
extern Body     *body_unlink(Prod *production, Body *curr_body, Body *prev_body);
extern void      body_free(Body *body);
/* String methods */
extern SItem    *string_copy(SItem *string, SItem **tail_ret);
extern SItem    *string_find_sitem(SItem *haystack, Symb *needle);
extern SItem    *string_find_sitem_from_set(SItem *haystack, SItem *needles);
extern int       string_cmp(SItem *str1, SItem *str2);
extern void      string_print(SItem *string, FILE *stream);
extern void      string_free(SItem *string);



#endif /* __GRAMMAR_H__ */
