#include <stdlib.h> /* malloc() */
#include <ctype.h>  /* isprint() */
#include <assert.h>

#include "grammar.h"
#include "tools.h"
#include "scanner.h"

extern char *progname;

static const char *errNontermAsTokenType="%s:%d: declaring nonterminal `%s' as a token type\n";
static const char *errStartIDRedefed="%s:%d: redefinition of start identifier\n";
static const char *errTokenTypeAsStart="%s:%d: cannot use token type `%s' as a start identifier\n";
static const char *errTokenTypeAsHead="%s:%d: cannot use token type `%s' as a production head\n";
static const char *errMemory       ="%s: error allocating memory\n";
static const char *errParse        ="%s:%d: parse error around `%c'\n";
static const char *errParseChar    ="%s:%d: parse error around character `\\x%x'\n";
static const char *errCharTTypeAsProdHead="%s:%d: cannot use character token type `%s' as a production head\n";
static const char *errNoRuleForStartID="%s: no rule for start identifier `%s'\n";
static const char *errUntermComm   ="%s:%d: unterminated comment\n";
static const char *errUntermProd   ="%s:%d: unterminated production\n";
static const char *warnStartIDUndef="%s: warning: no explicit start identifier definition, assuming\n\
    `%s' to be start identifier\n";



void *xmalloc(int size)
{
    void *ptr;

    if ( !(ptr = malloc(size)) ) {
        fprintf(stderr, errMemory, progname);
        exit(1);
    }

    return ptr;
}

/*
 * Grammar is read from stream_in.  Errors and warnings are reported to stderr
 * using global variable progname as program's name.  Doesn't check that the
 * grammar is non-empty.  Doesn't check that every nonterminal has a production
 * for it--such case would be eleminated by removing useless symbols
 */
Grammar *read_grammar(FILE *stream_in)
{
    Grammar *grammar;
    int scan_status;
    Prod *curr_prod = NULL;
    Body *curr_body = NULL;
    Symb *symb;

    assert(stream_in);
    yyin = stream_in;

    grammar  = grammar_new();
    grammar->alphabet = alphabet_new();

    /*
     * Main parse loop
     */
    while ( (scan_status = yylex()) ) {
        switch( scan_status ) {
        case SCANNED_NAMED_TOKEN_TYPE_DEF:
            if ( !alphabet_add_name(grammar->alphabet,
                                    1, yytext, yyleng, &symb) ) {
                /* This ID was used as a nonterminal, and now it is defined to
                 * be a token type--can't do */
                fprintf(stderr, errNontermAsTokenType, progname, yylineno, yytext);
                exit(1);
            }
            break;
        case SCANNED_START_ID_DEF:
            if ( grammar->start ) {   /* already defined... */
                if ( strcmp(grammar->start->name, yytext) ) {
                    /* ... with a different name */
                    fprintf(stderr, errStartIDRedefed, progname, yylineno);
                    exit(1);
                }
            } else {            /* none defined yet */
                if ( !alphabet_add_name(grammar->alphabet,
                                        0, yytext, yyleng, &grammar->start) ) {
                    /* Start ID can only be a nonterminal */
                    fprintf(stderr, errTokenTypeAsStart, progname, yylineno, yytext);
                    exit(1);
                }
            }
            break;
        case SCANNED_PRODUCTION_HEAD:
            if ( !alphabet_add_name(grammar->alphabet,
                                    0, yytext, yyleng, &symb) ) {
                /* Trying to use a token type as production head */
                fprintf(stderr, errTokenTypeAsHead, progname, yylineno, yytext);
                exit(1);
            }
            curr_prod = grammar_add_empty_production(grammar, symb);
            /* This will add an empty body (`epsilon-production') to this block
             * of productions.  If this is a real epsilon-production (next token
             * is a pipe or a semicolon), it will remain so; if not, then
             * symbols will be added to it, thus making it a non-epsilon
             * production */
            curr_body = production_add_empty_body(curr_prod);
            break;
        case SCANNED_PBODY_IDENTIFIER:
            /* If the identifier is not already in the alphabet, add it as a
             * nonterminal */
            if ( !(symb = alphabet_find_name(grammar->alphabet, yytext)) ) {
                int added;
                added = alphabet_add_name(grammar->alphabet,
                                          0, yytext, yyleng, &symb);
                assert(added);
            }
            /* Add it to the string of the current body */
            body_add_symb(curr_body, symb);
            break;
        case SCANNED_PBODY_CHAR_TOKEN_TYPE: {
            int added;
            added = alphabet_add_name(grammar->alphabet,
                                      1, yytext, yyleng, &symb);
            assert(added);
            body_add_symb(curr_body, symb);
            break; }
        case SCANNED_PBODY_PIPE:
            /* This will add an empty body (`epsilon-production') to this block
             * of productions.  If this is a real epsilon-production (next token
             * is a semicolon), it will remain so; if not, then symbols will be
             * added to it, thus making it a non-epsilon production */
            curr_body = production_add_empty_body(curr_prod);
            break;
        case SCANNED_PRODUCTION_END:
            curr_prod = NULL;
            break;
        case SCANNED_ERR_CHAR_TTYPE_AS_PRODUCTION_HEAD:
            fprintf(stderr, errCharTTypeAsProdHead, progname, yylineno, yytext);
            exit(1);
            break;
        case SCANNED_ERR_CHAR_TTYPE_AS_START:
            fprintf(stderr, errTokenTypeAsStart, progname, yylineno, yytext);
            exit(1);
            break;
        case SCANNED_ERR_UNTERM_COMMENT:
            fprintf(stderr, errUntermComm, progname, yylineno);
            exit(1);
            break;
        case SCANNED_ERR_UNTERM_PRODUCTION:
            fprintf(stderr, errUntermProd, progname, yylineno);
            exit(1);
            break;
        default: /* SCANNED_ERR */
            if ( isprint(*yytext) )
                fprintf(stderr, errParse, progname, yylineno, *yytext, yyleng);
            else
                fprintf(stderr, errParseChar, progname, yylineno, *yytext, yyleng);
            exit(1);
        }
    }

    /*
     * Sanity checks for start ID
     */
    if ( !grammar->start ) { /* no explicit start ID definition */
        /* Find first nonterminal, assume it to be start ID */
        grammar->start = grammar->alphabet->head;
        while ( grammar->start && grammar->start->terminal )
            grammar->start = grammar->start->next;
        /* grammar->start can still be NULL (if alphabet is empty, i.e. grammar
         * has no rules).  Sure a strange way to write grammars */
        if ( grammar->start )
            /* Print a warning */
            fprintf(stderr, warnStartIDUndef, progname, grammar->start->name);
    } else {
        /* We know start ID is a nonterminal (we were adding it to alphabet as
         * such), but make sure there is a rule with the start ID as a head */
        if ( !grammar_find_production_by_symb(grammar, grammar->start) ) {
            fprintf(stderr, errNoRuleForStartID, progname, grammar->start->name);
            exit(1);
        }
    }


    return grammar;
}



/*
 * Local variables:
 * compile-command: "make -k && ./unreachable < grammar.3"
 * End:
 */
