#include "grammar.h"
#include "tools.h"

#include <stdlib.h>
#include <assert.h>


/* Checks that the body is a unit production */
#define IS_UNIT_PRODUCTION(body)                                        \
  ( body->head                  &&   /* non-epsilon production... */    \
    !body->head->next           &&   /* ...consisting of a single... */ \
    !body->head->symb->terminal    ) /* ...nonterminal... */

static SItem *build_unit_pairs(Grammar *grammar, Symb *symb);
static Prod  *build_nonunit_prod(Grammar *grammar, Symb *hsymb, SItem *pairs);



/*
 * Returns a grammar built from the given grammar by removing unit productions
 */
Grammar *remove_units(Grammar *grammar)
{
    Grammar *grammar_out;
    Prod    *prod;

    assert(grammar);
    assert(grammar->alphabet);
    grammar_out           = grammar_new();
    grammar_out->start    = grammar->start;
    grammar_out->alphabet = grammar->alphabet;

    prod = grammar->head;
    while ( prod ) {
        SItem *unit_pairs = build_unit_pairs(grammar, prod->hsymb);
        Prod  *new_prod   = build_nonunit_prod(grammar, prod->hsymb, unit_pairs);

        if ( new_prod ) {
            if ( !grammar_out->head )
                grammar_out->head = new_prod;
            else {
                new_prod->next = grammar_out->head;
                grammar_out->head = new_prod;
            }
        }

        string_free(unit_pairs);
        prod = prod->next;
    }

    return grammar_out;
}



static SItem *build_unit_pairs(Grammar *grammar, Symb *symb)
{
    SItem *unit_pairs, *curr, *last;

    assert(grammar);
    assert(!symb->terminal);

    /* The symbol itself alway forms a unit pair with itself */
    curr = last = unit_pairs = (SItem *) xmalloc(sizeof(SItem));
    unit_pairs->symb = symb;
    unit_pairs->next = NULL;

    /* Now we are going to augment unit_pairs (set of symbols forming unit pairs
     * with symb symbol) with unit-pair forming symbols (from productions with
     * head symbols from unit_pairs) .  We will do so until we have checked
     * (productions for) each symbol in our set.  Note that we have to add new
     * unit-pair forming symbols to the end of the set so that they would get
     * checked later on */
    while ( curr ) {
        Prod *prod = grammar_find_production_by_symb(grammar, curr->symb);
        if ( prod ) { /* there is a production with this head (this should
                       * always be so if useless symbols had been removed) */
            /* Search thru the bodies for new (unit-pair forming) symbols */
            Body *body = prod->head;
            while ( body ) {
                if ( IS_UNIT_PRODUCTION(body) &&
                     !string_find_sitem(
                         unit_pairs, body->head->symb) ) { /* ...not yet in */
                    /* Add to the end */
                    last->next = (SItem *) xmalloc(sizeof(SItem));
                    last = last->next;
                    last->symb = body->head->symb;
                    last->next = NULL;
                }
                body = body->next;
            }
        }
        curr = curr->next;
    }

    return unit_pairs;
}



static Prod *build_nonunit_prod(Grammar *grammar, Symb *hsymb, SItem *pairs)
{
    Prod *prod_out;

    assert(grammar);
    assert(hsymb);
    assert(!hsymb->terminal);
    assert(pairs);

    prod_out = xmalloc(sizeof(Prod));
    prod_out->hsymb = hsymb;
    prod_out->next  = NULL;
    prod_out->head  = NULL;

    while ( pairs ) {
        Prod *prod = grammar_find_production_by_symb(grammar, pairs->symb);
        if ( prod ) {
            Body *body = prod->head;
            while ( body ) {
                if ( !IS_UNIT_PRODUCTION(body) &&
                     !production_find_body(prod_out, body) ) {
                    if ( !prod_out->head )
                        prod_out->head = body_copy(body);
                    else {
                        Body *next = prod_out->head;
                        prod_out->head = body_copy(body);
                        prod_out->head->next = next;
                    }
                }
                body = body->next;
            }
        }
        pairs = pairs->next;
    }

    if ( prod_out->head )
        return prod_out;

    free(prod_out);

    return NULL;
}



/*
 * Local variables:
 * compile-command: "make -k && ./units < grammar.4"
 * End:
 */
