#include "grammar.h"
#include "tools.h"
#include <stdlib.h>
#include <assert.h>



static SItem   *build_reachable_set(Grammar *grammar);
static SItem   *build_unreachable_set(Grammar *grammar, SItem *reachable);
static void     do_the_grammar(Grammar *grammar, SItem *unreachable);



/*
 * Removes from the given grammar rules containing unreachable symbols.  If
 * reachable_ret (unreachable_ret) is not NULL, stores there a pointer to a
 * string which is a set of reachable (unreachable) symbols from the grammar's
 * alphabet
 */
void remove_unreachable( Grammar *grammar,
                         SItem **reachable_ret, SItem **unreachable_ret )
{
    SItem *reachable, *unreachable;

    assert(grammar);
    assert(grammar->alphabet);

    /* Find all reachable terminals and nonterminals */
    reachable = build_reachable_set(grammar);

    /* Usually, number of reachable symbols is greater than that of unreachable.
     * Therefore it is usually more efficient to check symbols in productions
     * for belonging to the unreachable set rather than not belonging to the
     * reachable set.  Hence we build this unreachable symbols set */
    unreachable = build_unreachable_set(grammar, reachable);

    /* Build the new grammar */
    do_the_grammar(grammar, unreachable);

    /* Either return the reachable symbols set or free up all its items */
    if ( reachable_ret )
        *reachable_ret = reachable;
    else
        string_free(reachable);

    /* Either return the unreachable symbols set or free up all its items */
    if ( unreachable_ret )
        *unreachable_ret = unreachable;
    else
        string_free(unreachable);

    return;
}



static SItem *build_reachable_set(Grammar *grammar)
{
    SItem *reachable, *curr, *last;

    assert(grammar);
    assert(grammar->alphabet);
    assert(grammar->start);

    /* Start symbol is always reachable, so add it to the set */
    curr = last = reachable = (SItem *) xmalloc(sizeof(SItem));
    reachable->symb = grammar->start;
    reachable->next = NULL;

    /* Now we are going to augment the set with symbols which we encounter in
     * productions with head symbols we have already figured out to be
     * reachable.  We will do so until we have checked (productions for) each
     * symbol in our set.  Note that we have to add new reachable symbols to the
     * end of the set so that they would get checked later on */
    while ( curr ) {
        if ( !curr->symb->terminal ) { /* head symbols cannot be terminal */
            Prod *prod = grammar_find_production_by_symb(grammar, curr->symb);
            if ( prod ) { /* there is a production with this head */
                /* Search thru each of the bodies for new (reachable) symbols */
                Body *body = prod->head;
                while ( body ) {
                    SItem *sitem = body->head;
                    while ( sitem ) {
                        if ( !string_find_sitem(reachable, sitem->symb) ) {
                            /* Not in reachable set yet, add to the end */
                            last->next = (SItem *) xmalloc(sizeof(SItem));
                            last = last->next;
                            last->symb = sitem->symb;
                            last->next = NULL;
                        }
                        sitem = sitem->next;
                    }
                    body = body->next;
                }
            }
        }
        curr = curr->next;
    }

    return reachable;
}



static SItem *build_unreachable_set(Grammar *grammar, SItem *reachable)
{
    SItem *unreachable = NULL;
    Symb  *symb;

    assert(grammar);
    assert(grammar->alphabet);
    assert(reachable); /* should always be at least one reachable symbol */
    symb = grammar->alphabet->head;

    while ( symb ) {
        if ( !string_find_sitem(reachable, symb) ) {
            /* Symbol is not in reachable set, so add it to the head of
             * unreachable set list */
            SItem *next = NULL;
            if ( unreachable ) /* not empty */
                next = unreachable;
            unreachable = (SItem *) xmalloc(sizeof(SItem));
            unreachable->next = next;
            unreachable->symb = symb;
        }
        symb = symb->next;
    }

    return unreachable;
}



/*
 * Removes from the grammar those productions which contain an unreachable
 * symbol (either in their heads or their bodies)
 */
static void do_the_grammar(Grammar *grammar, SItem *unreachable)
{
    Prod *curr_prod, *prev_prod;

    assert(grammar);
    if ( !unreachable ) /* no unreachable symbols, grammar stays the same */
        return;

    curr_prod   = grammar->head;
    prev_prod   = NULL;
    while ( curr_prod ) {
        if ( string_find_sitem(unreachable, curr_prod->hsymb) )
            /* Head symbol of all bodies in this block is an unreachable symbol,
             * so dump the entire block */
            curr_prod = production_unlink(grammar, curr_prod, prev_prod);
        else {
            Body *curr_body = curr_prod->head, *prev_body = NULL;
            /* For each body, check that it contains no unreachable symbols */
            while ( curr_body ) {
                if ( curr_body->head &&
                     string_find_sitem_from_set(curr_body->head, unreachable) )
                    /* Unreachable symbol found, remove body */
                    curr_body = body_unlink(curr_prod, curr_body, prev_body);
                else {
                    /* This body is empty (epsilon-production) or no unreachable
                     * symbol found, let the body alone */
                    prev_body = curr_body;
                    curr_body = curr_body->next;
                }
            }
            if ( !curr_prod->head ) /* all bodies removed, remove production */
                curr_prod = production_unlink(grammar, curr_prod, prev_prod);
            else {                  /* skip over this production */
                prev_prod = curr_prod;
                curr_prod = curr_prod->next;
            }
        }
    }

    return;
}



/*
 * Local variables:
 * compile-command: "make -k && ./unreachable < grammar.3"
 * End:
 */
